; select the component with the highest certainty factor of a failing type to test
; it selects the component with the highest certainty when the type is not correct
; important is that both states are not equal to correct, but they can NOT be equal and also CAN be equal to eachother
(defrule selectComponentToTest
         (declare (salience 100))
         ?g <- (cfNode (gateNumber ?n) (state ?c) (certainty ?prob) (nodeType state))
         (not (cfNode (state ?c1&:(neq ?c1 correct)) (certainty ?qb2&:(> ?qb2 ?prob)) (nodeType state)))
         (test (neq ?c correct))
         (not (stop))
         =>
         (assert (testComponent ?n))
)

; store the results of measures
(deftemplate testResult
             (slot number
                   (type INTEGER)
             )
             (multislot inputs
                        (type INTEGER)
             )
             (slot output
                   (type INTEGER)
             )
             (multislot testProcessed
                   (type SYMBOL)
             )
)

; keeps track of how many and which cfNodes need to be given a new certainty factor
(deftemplate testReport
             (slot counter
                   (type INTEGER)
             )
             (multislot ids
                        (type INTEGER)
             )
)

; test the component on all possible in/outputs, and also execute the special tasks such as replacing or stopping or dooming
(defrule testComponent
         (declare (salience 110))
         ?gb   <- (gateBehaviour (gateType ?gt) (stateType ?c) (valueP ?ip) (valueQ ?iq) (output ?io))
         ?g    <- (gate (number ?n) (gate_type ?gt))
         (not (testResult (number ?n) (inputs $?i&:(eq $?i (create$ ?ip ?iq)))))
         ?test <- (testComponent ?n)
         (not (stop))
         =>
         (if (neq ?iq -20)
             then (bind ?response (askTest (str-cat "Test component number " ?n ", set the first input to " ?ip " and the second input to "
               ?iq ". What is the output?")))
             else (bind ?response (askTest (str-cat "Test component number " ?n ", set the input to " ?ip ". What is the output?")))
         )
         (retract ?test)
         (if (or (eq ?response "0") (eq ?response "1"))
             then (assert (testResult (number ?n) (inputs ?ip ?iq) (output (integer (string-to-field ?response)))))
                  (assert (testReport (counter 0) (ids )))
                  (printout t "Processing.." crlf)
             else (assert (testComponent ?n)) ; necessary otherwise afther this the same testing won't start again
                  (if (eq ?response "replace")
                      then (bind ?removeGate (ask "Which gate do you want to replace?"))
                           (do-for-fact
                                       ((?cf cfNode))
                                       (eq ?cf:id (integer (string-to-field ?removeGate)))
                                       (assert (removeCfNode (integer (string-to-field ?removeGate))))
                           )
                      else (if (eq ?response "stop")
                           then (assert (stop))
                           else (bind ?doomGate (ask "Which gate do you want to doom?"))
                                (do-for-fact
                                       ((?cf cfNode))
                                       (and (eq ?cf:gateNumber (integer (string-to-field ?doomGate))) (member$ gut ?cf:nodeType))
                                       (modify ?cf (certainty 0.8) (update TRUE))
)
                           )
                   )
         )
)


; propagate the test result through the model, matches on a testresult and the gate+behaviour and then adds it to the testreport,
; or adds a remove fact if the output doesn't match, so this state can be removed
(defrule processTestResult
         (declare (salience 120))
         ?test <- (testResult (number ?n) (inputs $?inp ) (output ?response) (testProcessed $?tested))
         ?gs   <- (gate (number ?n) (gate_type ?gt))
         ?gb   <- (gateBehaviour (gateType ?gt) (stateType ?c) (valueP ?ip) (valueQ ?iq) (output ?io))
         ?cfN  <- (cfNode (id ?id) (dependent $?dep) (certainty ?cert) (gateNumber ?n) (state ?c) (nodeType $?nType) (update FALSE))
         ?tr <- (testReport (counter ?count) (ids $?ids))
         (test (eq $?inp (create$ ?ip ?iq)))
         (test (eq $?nType (create$ test ?ip ?iq)))
         (test (not (member$ ?c $?tested)))
         =>
         (if (eq (integer ?response) (integer ?io))
            then (modify ?tr (counter (+ ?count 1)) (ids (insert$ $?ids 1 ?id)))
            else (assert (removeCfNode ?n ?c))
         )
         (modify ?test (testProcessed (insert$ $?tested 1 ?c)))
)

; loops over the ids given in the testReport and updates the certainty factor, and sets the update on true, so the network will be propagated
(defrule processTestReport
         (declare (salience 119))
         ?t <- (testReport (counter ?c) (ids $?i))
         =>
         (loop-for-count (?teller 1 ?c) do
                         (do-for-fact
                                   ((?cf cfNode))
                                   (member$ ?cf:id $?i)
                                   (modify ?cf (certainty (/ 1.0 (float ?c))) (update TRUE))
                         )
         )
         (retract ?t)
)

; alerts the user if some defect rises above a threshold
(defrule alertDefectNode
         (declare (salience 700))
         ?cfn <- (cfNode (gateNumber ?n) (state ?c) (certainty ?cert) (nodeType state))
         (gateThreshold ?thres)
         (test (> ?cert ?thres))
         (test (neq ?c correct))
         =>
         (printout t "Gate number " ?n " is probably " ?c ". You should replace it." crlf)
         (retract ?cfn)
         (bind ?response (ask "Is the circuit still broken? Type (Y) to continue, (N) to stop."))
         (if (eq (upcase ?response) "N")
             then (assert (stop))
         )
)

; alerts the user about a wrong gate (known for certain!), and asks to proceed or not
(defrule alertNodeDetermined
         (declare (salience 701))
         ?cfn <- (cfNode (gateNumber ?n) (state ?c) (nodeType state))
         (not (cfNode (gateNumber ?n) (state ?c1&:(neq ?c1 ?c)) (nodeType state)))
         (test (neq ?c correct))
         =>
         (printout t "We have concluded the gate with number " ?n " has state: " ?c crlf)
         (printout t "Please replace it for a new one." crlf)
         (retract ?cfn)
         (bind ?response (ask "Is the circuit still broken? Type (Y) to continue, (N) to stop."))
         (if (eq (upcase ?response) "N")
             then (assert (stop))
         )
)

; warns the user that there are no components that can be broken
(defrule alertNoNodes
         (declare (salience 700))
         (not (initial-fact))
         (not (cfNode (state ?c1&:(neq ?c1 correct))))
         =>
         (printout t "All components seem allright, so the person who brought this should go up the roof."crlf)
)

; this removes all cfNodes for a certain state of a gate, that state has been proven not the case by a testresult.
(defrule removecfNodes
         (declare (salience 750))
         ?remo <- (removeCfNode ?n ?c)
         ?cfn <- (cfNode (gateNumber ?n) (state ?c))
         =>
         (retract ?cfn)
)

; remove the testComponent fact in case the current gate in testing has been chosen to replace
(defrule removeTestcomponent
         (declare (salience 751))
         ?remo <- (removeCfNode ?n)
         ?t <- (testComponent ?n)
         =>
         (retract ?t)
)

; remove all the cfNodes of one gate (in the case of replace)
(defrule removecfNodesGate
         (declare (salience 751))
         ?remo <- (removeCfNode ?n)
         ?cfn <- (cfNode (gateNumber ?n))
         =>
         (retract ?cfn)
)

; remove the remove fact (only for cleaning up facts purpose)
(defrule removeRemovers
          (declare (salience 740))
          ?remo <- (removeCfNode ?n ?c)
          (not (cfNode (gateNumber ?n) (state ?c)))
          =>
          (retract ?remo)
)

; remove the remove fact (only for cleaning up facts purpose)
(defrule removeRemovers2
          (declare (salience 740))
          ?remo <- (removeCfNode ?n)
          (not (cfNode (gateNumber ?n)))
          =>
          (retract ?remo)
)

