; the definition of the nodes in the certainty factor network
(deftemplate cfNode
             (slot id
                   (type INTEGER)
             )
             (multislot dependent
                   (type INTEGER)
             )
             (slot certainty
                   (type FLOAT)
             )
             (slot gateNumber
                   (type INTEGER)
             )
             (slot state
                   (type SYMBOL)
             )
             (multislot nodeType
                   (type SYMBOL)
             )
             (slot update
                   (type SYMBOL)
             )
)

; this function calculates the coconlude rule. if it forces to go in recursion, than 
; it checks if both certainties are zero, to avoid infinite recursionty
(deffunction coconclude (?a ?b)
             (if (and (> ?a 0) (> ?b 0))
                 then (return (+ ?a (* ?b (- 1 ?a))))
                 else (if (or (> ?a 0) (> ?b 0))
                          then (return (/ (+ ?a ?b) (- 1 (min (abs ?a) (abs ?b)))))
                          else (if (and (eq ?a 0) (eq ?b 0))
                                   then (return 0)
                                   else (return (* -1 (coconclude (* -1 ?a) (* -1 ?b))))
                               )
                      )
             )
)

; make sure that after the parsing is done, these propagating rule fire first before 
; and directly after the nodes are adjusted by tests of the user
(defrule cfCoconclude
         (declare (salience 800))
         ?cf1 <- (cfNode (dependent $?dep1) (certainty ?cert1) )
         ?cf2 <- (cfNode (id ?id2) (certainty ?cert2) (update ?u1))
         ?cf3 <- (cfNode (id ?id3) (certainty ?cert3) (update ?u2))
         (test (eq $?dep1 (create$ ?id2 ?id3)))
         (test (or (eq ?u1 TRUE) (eq ?u2 TRUE)))
         =>
         (bind ?newCert (coconclude ?cert2 ?cert3))
         (modify ?cf1 (certainty ?newCert) (update TRUE))
         (modify ?cf2 (update FALSE))
         (modify ?cf3 (update FALSE))
)

; starting point of the program, to load the circuit to start reasoning
; it also starts the cf id counter to aid other rules in enforcing the
; invariant that every cfNode has a unique id
(defrule start
         ?i <- (initial-fact)
         =>
         (retract ?i)
         (assert (cfIdCounter 0))
)

; propagate the a priori probabilities through the network by creating the network
; it basically distinguishes between gates with 2 or 4 inputs and builds the
; certainty factor networks as described in the appended report, while mainting
; the invariant that every cfNode has a unique id
(defrule makeCfNetwork
         (declare (salience 900))
         ?f       <- (gate (number ?n) (gate_type ?gt))
         ?prior   <- (gatePrioriProb (type ?gt)  (state ?c) (prob ?prob))
         ?cfCount <- (cfIdCounter ?count)
         (not (gateParsed ?n ?c))
         =>
         (assert (gateParsed ?n ?c))
         (retract ?cfCount)
         (if (and (neq ?gt input) (neq ?gt output))
             then (if (or (eq ?gt and) (eq ?gt or) (eq ?gt xor))
                   then (assert (cfNode (id ?count) (gateNumber ?n) (certainty ?prob) (state ?c) (nodeType priori) (update FALSE)) )
                         (assert (cfNode (id (+ 1 ?count)) (gateNumber ?n) (state ?c) (certainty 0.0) (nodeType (create$ test 0 0))(update FALSE)))
                         (assert (cfNode (id (+ 2 ?count)) (gateNumber ?n) (state ?c) (certainty 0.0) (nodeType (create$ test 1 0)) (update FALSE)))
                         (assert (cfNode (id (+ 3 ?count)) (gateNumber ?n) (state ?c) (certainty 0.0) (nodeType (create$ test 0 1)) (update FALSE)))
                         (assert (cfNode (id (+ 4 ?count)) (gateNumber ?n) (state ?c) (certainty 0.0) (nodeType (create$ test 1 1)) (update FALSE)))
                         (assert (cfNode (id (+ 5 ?count)) (gateNumber ?n) (state ?c) (certainty 0.0) (nodeType gut) (update FALSE)))
                         (assert (cfNode (id (+ 6 ?count)) (gateNumber ?n) (certainty ?prob)(state ?c) (dependent ?count (+ 1 ?count)) (update FALSE)))
                         (assert (cfNode (id (+ 7 ?count)) (gateNumber ?n) (state ?c) (dependent (+ 2 ?count) (+ 3 ?count)) (update FALSE)))
                         (assert (cfNode (id (+ 8 ?count)) (gateNumber ?n) (state ?c) (dependent (+ 4 ?count) (+ 5 ?count)) (update FALSE)))
                         (assert (cfNode (id (+ 9 ?count)) (gateNumber ?n) (state ?c) (dependent (+ 7 ?count) (+ 8 ?count)) (update FALSE)))
                         (assert (cfNode (id (+ 10 ?count)) (gateNumber ?n) (certainty ?prob) (state ?c) (dependent (+ 6 ?count) (+ 9 ?count)) (nodeType state) (update FALSE)))
                         (assert (cfIdCounter (+ ?count 11)))
                   else (assert (cfNode (id ?count) (gateNumber ?n) (certainty ?prob) (nodeType priori) (update FALSE)))
                         (assert (cfNode (id (+ 1 ?count)) (gateNumber ?n) (state ?c) (certainty 0.0) (nodeType (create$ test 0)) (update FALSE)))
                         (assert (cfNode (id (+ 2 ?count)) (gateNumber ?n) (state ?c) (certainty 0.0) (nodeType (create$ test 1)) (update FALSE)))
                         (assert (cfNode (id (+ 3 ?count)) (gateNumber ?n) (state ?c) (certainty 0.0) (nodeType gut) (update FALSE)))
                         (assert (cfNode (id (+ 4 ?count)) (gateNumber ?n) (certainty ?prob) (state ?c) (dependent ?count (+ 1 ?count)) (update FALSE)))
                         (assert (cfNode (id (+ 5 ?count)) (gateNumber ?n) (state ?c) (dependent (+ 2 ?count) (+ 3 ?count)) (update FALSE)))
                         (assert (cfNode (id (+ 6 ?count)) (gateNumber ?n) (certainty ?prob) (state ?c) (dependent (+ 4 ?count) (+ 5 ?count)) (nodeType state) (update FALSE)))
                         (assert (cfIdCounter (+ ?count 5)))
                   )
         )
)
; Template to keep track of which number the gates is at
; and also to keep track of which gates also have been used to created line gates
(deftemplate lineGatesAdded
             (slot gatenumber
                   (type INTEGER)
             )
             (multislot ids
                   (type INTEGER)
             )
)

; Look up the highest gatenumber available
(defrule gateNumberCreate
         (declare (salience 1100))
         ?gate <- (gate (number ?n))
         (not (gate (number ?n1&:(> ?n1 ?n))))
         (not (lineGatesAdded))
         (parseLines TRUE)
         =>
         (assert (lineGatesAdded (gatenumber ?n)))
)


; Use the outputs of a gate to create linegates that go from the gate itself to the output
(defrule createLinegates
         (declare (salience 1000))
         ?linegates <- (lineGatesAdded (gatenumber ?num) (ids $?ids))
         ?gate1 <- (gate (number ?n) (gate_type ?t) (outputs $?outp))
         (not (test (member$ ?n $?ids)))
         (not (test (eq ?t line)))
         =>
         (modify ?linegates (gatenumber (+ ?num (length$ $?outp))) (ids (insert$ $?ids 1 ?n)))
         (loop-for-count (?teller 1 (length$ $?outp)) do
                         (assert (gate (number (+ ?num ?teller)) (gate_type line) (inputs (create$ ?n)) (outputs (create$ (nth$ ?teller $?outp)))))

         )
 )